--	SMart Heal (Druid)
--	Written by scott@mithrandir.com
--	Scott's attempt to do something initially based on DruidPartyHeal and Healomatic
--	but with inspiration and code derived from many.

--	Notes
--
--	This does not use Tranquility.  It only uses Healing Touch, Regrowth and Rejuvenation.
--
--	There is no way to determine LOS on the healee.  We can determine whether the healee is in range
--	but if not in LOS we will get an error message when attempting to cast.
--
--	There is no API that detects movement.  Therefore, we attempt an approximation using the
--	GetPlayerMapPosition() call to record X, Y coords for the character to guess when moving.
--	This system does not work well, so we sometimes attempt to cast something other than the
--	Rejuvenation spell.
--
--	There have been problems ascertaining whether Nature's Swiftness is castable, i.e., not in
--	cooldown.  This is because GetActionCooldown() has been quite unreliable.  Therefore, this
--	ability is not turned on by default since an attempt at casting this will not try a heal
--	and it can fail to cast because it is still in cooldown.
--
--	TODO
--
--	Our code makes use of the algorithm "data structure" but we need a way for the user to define
--	their own algorithm.
--
--	Just as we have four priority levels for character classes, we should have the same for
--	characters.  We need to set up the system to make use of a "data structure" and also set up
--	a way for users to be able to assign specific characters to priority levels.
--

SMH_NAME = "SMart Heal (Druid)";
SMH_VERSION = "0.9.4";
SMH_VERSION_STRING = SMH_NAME.." "..SMH_VERSION;
SMH_MACRO_COMMAND = "/smh";

local _SMH_class;				-- class of the player.  used to determine whether we should enable the addon.
local _SMH_myName;				-- name of the player.  only use in determining raid group.
local _SMH_event_dispatch = {};			-- table of functions whose keys are the events
local _SMH_lastX = nil;				-- where we last were in X
local _SMH_lastY = nil;				-- where we last were in Y
local _SMH_isMoving = false;			-- determined using above two variables
local _SMH_healingSlot = nil;			-- indicates the slot in action bar that has one of the healing spells in which we have interest -- for range detection
local _SMH_swiftNatureSlot = nil;		-- indicates the slot in action bar that has Nature's Swiftness -- so we can check for cooldown for automatic casting
local _SMH_extraHealingAmount = 0;		-- amount of extra healing that is done by equipment

SMH_CLASS_DRUID_ENGLISH = 'DRUID';		-- the non-localized version used to determine if we are a Druid

--	Two variables are used.  The first is the set of defaults that are always present in the system.
--	These are the values that are used when the user resets the options to default status.  This is
--	also used to ensure the user's stored variables have the proper keys in them.  The second variable
--	are the actual option values that the user sets.  These are persistent between runs of the program
--	and work across versions of this addon because we check to ensure old keys are removed and new
--	keys are added after the stored variable is loaded.
SMH_Configuration_Defaults = {
	["messagesParty"] = true,		-- controls whether the spell cast messages appear in the party chat window
	["debug"] = false,			-- indicates whether we are in debug mode
	["targetPriority"] = true,		-- indicates whether the "target" has priority over all else
	["healPercentage"] = 50,		-- indicates the percentage of the possible heal range to use to compare to the target's health
	["healingPets"] = true,			-- indicates whether party pets are considered in the healing list
	["checkingNaturesSwiftness"] = false,	-- indicates whether we think about casting Nature's Swiftness automatically as part of heal process
	["threshold"] = 100,			-- indicates percentage of health target must be below to be considered for healing
	["petThreshold"] = 100,			-- indicates percentage of health pets must be below to be considered for healing
	["selfThreshold"] = 40,			-- indicates percentage of health self must be below to be forced to head of healing list
	["healingRaid"] = true,			-- indicates whether raid members are to be healed
	["healingRaidPets"] = true,		-- indicates whether raid pets are to be healed
	["partyOverRaid"] = true,		-- indicates whether the party has priority over the raid for healing
	["healsParty"] = true,			-- indicates whether parts members other than self are healed
	["algorithm"] = {
		["75"] = SMH_SPELL_REJUVENATION,
		["50"] = SMH_SPELL_REGROWTH,
		["25"] = SMH_SPELL_HEALING_TOUCH,
		["0"] = SMH_SPELL_REGROWTH,
		},				-- indicates the spells that are used in healing percentage ranges
	["defaultSpells"] = { 
		SMH_SPELL_REJUVENATION,
		SMH_SPELL_REGROWTH,
		SMH_SPELL_HEALING_TOUCH,
		},				-- indicates the order of the spells that are attempted if none determined from algorithm
	["minimapButtonAngle"] = 356,		-- indicates the angle around the minimap for the location of the button that toggles options UI
	["priorityRaidGroup1"] = false,		-- indicates members of raid group 1 are given priority
	["priorityRaidGroup2"] = false,		-- indicates members of raid group 2 are given priority
	["priorityRaidGroup3"] = false,		-- indicates members of raid group 3 are given priority
	["priorityRaidGroup4"] = false,		-- indicates members of raid group 4 are given priority
	["priorityRaidGroup5"] = false,		-- indicates members of raid group 5 are given priority
	["priorityRaidGroup6"] = false,		-- indicates members of raid group 6 are given priority
	["priorityRaidGroup7"] = false,		-- indicates members of raid group 7 are given priority
	["priorityRaidGroup8"] = false,		-- indicates members of raid group 8 are given priority
	["priorityRaidPets"] = false,		-- indicates that pets of groups that have priority get the same priority
	["overhealRankHealingTouch"] = 0,	-- indicates rank of Healing Touch desired for overhealing
	["overhealRankRegrowth"] = 0,		-- indicates rank of Regrowth desired for overhealing
	["overhealRankRejuvenation"] = 0,	-- indicates rank of Rejuvenation desired for overhealing
	["classPriorities"] = {
		[SMH_CLASS_DRUID] = 2,
		[SMH_CLASS_HUNTER] = 2,
		[SMH_CLASS_MAGE] = 2,
		[SMH_CLASS_PALADIN] = 2,
		[SMH_CLASS_PRIEST] = 2,
		[SMH_CLASS_ROGUE] = 2,
		[SMH_CLASS_SHAMAN] = 2,
		[SMH_CLASS_WARLOCK] = 2,
		[SMH_CLASS_WARRIOR] = 2,
		};				-- indicates the priority that classes have
	["classPriorityOverParty"] = false,	-- indicates high priority classes have higher priority than party members
	["testingNewEquipmentParsing"] = true,	-- testing code:  changed multiple calls to walking through an array of patterns with same call structure code
	};

SMH_Configuration = {};

local outputBooleanMappings = { [false] = SMH_OUTPUT_OPTION_OFF, [true] = SMH_OUTPUT_OPTION_ON };

--	The cached spell book data is a two dimensional table with three rows representing
--	each of the spells and ten columns representing the ranks.  At each point in the table
--	is the index into the spell book for that spell.
local cachedSpellBookData = { };

--	Note that levelLearned cannot be used for determining whether a target can receive
--	a spell because Healing Touch can be cast on anyone of any level it seems.
local cachedSpellTargetLevels = {
	[SMH_SPELL_HEALING_TOUCH] = { 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
	[SMH_SPELL_REGROWTH] = { 12, 18, 24, 30, 36, 42, 48, 54, 60 },
	[SMH_SPELL_REJUVENATION] = { 4, 10, 16, 22, 28, 34, 40, 46, 52, 58 }
	};
local levelLearned = {
	[SMH_SPELL_HEALING_TOUCH] = { 1, 8, 14, 20, 26, 32, 38, 44, 50, 56 },
	[SMH_SPELL_REGROWTH] = { 12, 18, 24, 30, 36, 42, 48, 54, 60 },
	[SMH_SPELL_REJUVENATION] = { 4, 10, 16, 22, 28, 34, 40, 46, 52, 58 }
	};

local equipmentSlots = { "Head", "Neck", "Shoulder", "Back", "Chest", "Shirt", "Tabard", "Wrist", "Hands", "Waist", "Legs", "Feet", "Finger0", "Finger1", "Trinket0", "Trinket1", "MainHand", "SecondaryHand", "Ranged" };

local playerIds = { party1 = 1, party2 = 2, party3 = 3, party4 = 4 };
local petIds = { pet = 0, partypet1 = 1, partypet2 = 2, partypet3 = 3, partypet4 = 4 };

local function outputCombatMessage(message)
	if ChatFrame2 then
		ChatFrame2:AddMessage(message)
	end
end

local function outputChatMessage(message)
	if DEFAULT_CHAT_FRAME then
		DEFAULT_CHAT_FRAME:AddMessage(message)
	end
end

local function outputDebugMessage(message)
	if SMH_Configuration.debug then
		outputChatMessage(message)
	end
end

--	These functions and their use taken from Healomatic
local function parseInteger(text, index)
	local i = 1;
	for integer in string.gfind(text, '%d+') do
		if i == index then
			return tonumber(integer);
		end
		i = i + 1;
	end
end
local function piFirst(text)	return parseInteger(text, 1);	end
local function piFirstThird(text)	return parseInteger(text, 1) + parseInteger(text, 3);	end
local function piSecond(text)	return parseInteger(text, 2);	end
local function piSecondThird(text)	return parseInteger(text, 2) + parseInteger(text, 3);	end
--	Mapping of the various healing spells to parsers to get their amounts
local spellParsersMaxHeal = {
	[SMH_CLASS_DRUID_ENGLISH] = { [SMH_SPELL_HEALING_TOUCH] = piSecond, [SMH_SPELL_REJUVENATION] = piFirst, [SMH_SPELL_REGROWTH] = piSecondThird, ['Tranquility'] = piFirst },
	};
local spellParsersMinHeal = {
	[SMH_CLASS_DRUID_ENGLISH] = { [SMH_SPELL_HEALING_TOUCH] = piFirst, [SMH_SPELL_REJUVENATION] = piFirst, [SMH_SPELL_REGROWTH] = piFirstThird, ['Tranquility'] = piFirst }
	};

local function UnitIsHealable(unit)
	if not UnitExists(unit) then return false end
	if not UnitIsFriend('player', unit) then return false end
	if     UnitIsEnemy(unit, 'player') then return false end
	if     UnitCanAttack(unit, 'player') then return false end
	if     UnitCanAttack('player', unit) then return false end
	if     unit ~= 'player' and not UnitCanCooperate('player', unit) then return false end
	if not UnitIsConnected(unit) then return false end
	if     UnitIsDead(unit) then return false end
	if     UnitIsGhost(unit) then return false end
	return true
end

local function healthCompare(unit1, unit2)
	return unit1.percent < unit2.percent;
end

local function greaterCompare(one, two)
	return one > two;
end

local function hasBuff(unitName, buffName)
	local retval = false;
	local texture;
	for i=0,15,1 do texture = UnitBuff(unitName, i);
		if texture then
			smhTooltip:SetUnitBuff(unitName, i);
			if smhTooltipTextLeft1:GetText() == buffName then
				retval = true;
				outputDebugMessage(unitName.." has buff "..buffName);
				break;
			end
		end
	end
	return retval;
end

--	This returns a table containing keys of the buff names.
local function unitBuffs(unitName)
	local retval = { };
	for i = 0, 15, 1 do
		texture = UnitBuff(unitName, i);
		if texture then
			smhTooltip:SetUnitBuff(unitName, i);
			retval[smhTooltipTextLeft1:GetText()] = true;
		end
	end
	return retval;
end

local function findHealingSlot()
	outputDebugMessage("==> Finding healing slot");
	_SMH_healingSlot = nil;
	_SMH_swiftNatureSlot = nil;
	for i = 1, 120 do
		if (HasAction(i)) then
			smhTooltip:SetAction(i);
			local slotName = smhTooltipTextLeft1:GetText();
			if (slotName == SMH_SPELL_REGROWTH or slotName == SMH_SPELL_HEALING_TOUCH or slotName == SMH_SPELL_REJUVENATION) then
				_SMH_healingSlot = i;
			elseif (slotName == SMH_SPELL_NATURESSWIFTNESS) then
				_SMH_swiftNatureSlot = i;
			end
		end
		if (_SMMH_healingSlot ~= nil and _SMH_swiftNatureSlot ~= nil) then break end
	end
	if (_SMH_healingSlot ~= nil) then outputDebugMessage("Healing slot is ".._SMH_healingSlot); else outputDebugMessage("!!! Did Not find healing slot!!!"); end;
	if (_SMH_swiftNatureSlot ~= nil) then outputDebugMessage("Nat. Swift. slot is ".. _SMH_swiftNatureSlot); else outputDebugMessage("!!! Did Not find Nat. Swift. slot!!!"); end;
end

local function myRaidGroup()
	local myGroup = 0;
	local numberInRaid = GetNumRaidMembers();

	for i = 1, numberInRaid do
		local raidName, _, raidGroup = GetRaidRosterInfo(i);
		if (raidName == _SMH_myName) then
			myGroup = raidGroup;
			break;
		end
	end
	return myGroup;
end

local function tryToInsertUnitForHealing(tableName, unitName, parameterName, debugMessage)
	local currentHealth = UnitHealth(unitName);
	local maxHealth = UnitHealthMax(unitName);
	local percentHealth = currentHealth / maxHealth;
	-- Only put damaged units that meet the damage threshold into the table.
	if (currentHealth < maxHealth and percentHealth < (SMH_Configuration[parameterName] / 100.0)) then
		table.insert(tableName, 1, { id = unitName, percent = percentHealth, current = currentHealth, max = maxHealth });
		if (debugMessage ~= nil) then outputDebugMessage(debugMessage) end
	end
end

function SMH_algorithmString()
	local retval = SMH_TITLE_ALGORITHM;
	local temp = { };
	retval = retval .. "\n";
	for n in pairs(SMH_Configuration.algorithm) do table.insert(temp, n) end
	table.sort(temp, greaterCompare);
	for i, n in ipairs(temp) do 
		retval = retval..string.format("   %s%% %s\n", n, SMH_Configuration.algorithm[n])
	end
	com_mithrandir_smhAlgorithmDesc:SetText(retval);
end

-- TODO: Prepare this for localization
function SMH_priorityString()
	local retval = "Heal Priority List\n   1. Player (with Alt pressed)\n";
	local order = 2;
	local wordsToAdd;
	local tableSize;
	local priority = { [0] = { }, [1] = { }, [2] = { }, [3] = { } };

	-- We now take each of the class priorities and put them into the priority table.
	local temp = { };
	for n in pairs(SMH_Configuration.classPriorities) do table.insert(temp, n) end
	table.sort(temp, greaterCompare);
	for i, n in ipairs(temp) do
		table.insert(priority[SMH_Configuration.classPriorities[n]], n);
	end

	if SMH_Configuration.targetPriority then
		retval = retval.."   "..order..". Target\n";
		order = order + 1;
	end

	retval = retval.."   "..order..". Player with less than "..SMH_Configuration.selfThreshold.."% health\n";
	order = order + 1;

	if SMH_Configuration.classPriorityOverParty and table.getn(priority[3]) > 0 then
		retval = retval .. "   " .. order .. ". ";
		for i = 1, table.getn(priority[3]), 1 do
			if i ~= 1 then
				retval = retval .. "/";
			end
			retval = retval .. priority[3][i];
		end
		retval = retval .. "\n";
		order = order + 1;
	end

	if SMH_Configuration.partyOverRaid then
		retval = retval.."   "..order..". Player";
		if SMH_Configuration.healsParty then
			retval = retval.."/Party";
		end
		if SMH_Configuration.healingPets then
			retval = retval.."/Party Pets";
		end
		retval = retval.."\n";
		order = order + 1;
	end

	if not SMH_Configuration.classPriorityOverParty and table.getn(priority[3]) > 0 then
		local code = "";
		if SMH_Configuration.partyOverRaid then code = "[R]" end
		retval = retval .. "   " .. order .. ". ";
		for i = 1, table.getn(priority[3]), 1 do
			if i ~= 1 then
				retval = retval .. "/";
			end
			retval = retval .. priority[3][i] .. code;
		end
		retval = retval .. "\n";
		order = order + 1;
	end

	wordsToAdd = { };
	for i = 1, 8, 1 do
		local key = "priorityRaidGroup"..i;
		if SMH_Configuration[key] then
			table.insert(wordsToAdd, "RG"..i);
			if SMH_Configuration.priorityRaidPets then
				table.insert(wordsToAdd, "RG"..i.." Pets");
			end
		end
	end
	tableSize = table.getn(wordsToAdd);
	if tableSize > 0 then
		retval = retval.."   "..order..". ";
		for i = 1, tableSize, 1 do
			if (i ~= 1) then
				retval = retval.."/";
			end
			retval = retval..wordsToAdd[i];
		end
		retval = retval.."\n";
		order = order + 1;
	end

	wordsToAdd = { };
	if SMH_Configuration.healingRaid then
		table.insert(wordsToAdd, "Raid");
	end
	if SMH_Configuration.healingRaidPets then
		table.insert(wordsToAdd, "Raid Pets");
	end
	if not SMH_Configuration.partyOverRaid then
		table.insert(wordsToAdd, "Player");
		if SMH_Configuration.healsParty then
			table.insert(wordsToAdd, "Party");
		end
		if SMH_Configuration.healingPets then
			table.insert(wordsToAdd, "Party Pets");
		end
	end
	tableSize = table.getn(wordsToAdd);
	if tableSize > 0 then
		retval = retval.."   "..order..". ";
		for i = 1, tableSize, 1 do
			if (i ~= 1) then
				retval = retval.."/";
			end
			retval = retval..wordsToAdd[i];
		end
		retval = retval.."\n";
		order = order + 1;
	end

	if table.getn(priority[1]) > 0 then
		retval = retval .. "   " .. order .. ". ";
		for i = 1, table.getn(priority[1]), 1 do
			if i ~= 1 then
				retval = retval .. "/";
			end
			retval = retval .. priority[1][i];
		end
		retval = retval .. "\n";
		order = order + 1;
	end

	if table.getn(priority[0]) > 0 then
		retval = retval .. "\nClasses Not Healed\n";
		for i = 1, table.getn(priority[0]), 1 do
			if i ~= 1 then
				retval = retval .. "/";
			end
			retval = retval .. priority[0][i];
		end
		retval = retval .. "\n";
	end

	com_mithrandir_smhPriorityDesc:SetText(retval);
end

local function whichPriorityLevel(classModifier, isParty)
	local retval = classModifier;

	if isParty then
		if classModifier == 3 then
			if SMH_Configuration.classPriorityOverParty then
				retval = 6;
			elseif SMH_Configuration.partyOverRaid then
				retval = 5;
			else
				retval = 4;
			end
		elseif classModifier == 2 then
			if SMH_Configuration.partyOverRaid then
				retval = 5;
			end
		end
	else
		if classModifier == 3 then
			if SMH_Configuration.classPriorityOverParty then
				retval = 6;
			else
				retval = 4;
			end
		end
	end

	return retval;
end

local maxClassHealth = { [SMH_CLASS_DRUID] = 3000, [SMH_CLASS_HUNTER] = 3000, [SMH_CLASS_MAGE] = 2000, [SMH_CLASS_PALADIN] = 4000, [SMH_CLASS_PRIEST] = 2000, [SMH_CLASS_ROGUE] = 3000, [SMH_CLASS_SHAMAN] = 3000, [SMH_CLASS_WARLOCK] = 2000, [SMH_CLASS_WARRIOR] = 4000 };

local function orderedHealableUnits()
	local groupStatus = { };	-- the real results
	local tableSize;
	local tableBase;
	local classModifier;
	-- Note that we consider priority[6] higher than priority[1]
	local priority = { };		-- a table of tables
	local numberPriorities = 6;

	-- initialize our priority tables to empty
	for i = 1, numberPriorities, 1 do
		priority[i] = { };
	end

outputCombatMessage("SMart Heal Information Dump:");
outputCombatMessage(com_mithrandir_smhPriorityDesc:GetText());
	-- We now consider ourselves (as part of our own party)
	classModifier = SMH_Configuration.classPriorities[UnitClass("player")];
	if classModifier ~= 0 then
		tableBase = whichPriorityLevel(classModifier, true);
		tryToInsertUnitForHealing(priority[tableBase], "player", "threshold", nil);
	end

	if SMH_Configuration.healsParty then
		for i in playerIds do
			if UnitIsHealable(i) then
				classModifier = SMH_Configuration.classPriorities[UnitClass(i)];
				if classModifier ~= 0 then
					tableBase = whichPriorityLevel(classModifier, true);
					tryToInsertUnitForHealing(priority[tableBase], i, "threshold", nil);
				end
			end
		end
	end

	if SMH_Configuration.healingPets then
		for i in petIds do
			if UnitIsHealable(i) then
				if SMH_Configuration.partyOverRaid then tableBase = 5 else tableBase = 2 end
				tryToInsertUnitForHealing(priority[tableBase], i, "petThreshold", nil);
			end
		end
	end

	-- Determine if any of the raid subgroups have priority
	local priorityRaid = false;
	for i = 1, 8, 1 do
		if SMH_Configuration["priorityRaidGroup"..i] then
			priorityRaid = true;
		end
	end

	-- Only process the raid code if we need to do so.
	-- Note that since we do not know what subgroup the player is in we
	-- first record each of the damaged ones into subgroups because we
	-- will ignore the player's subgroup if we can.
	if SMH_Configuration.healingRaid or SMH_Configuration.healingRaidPets or priorityRaid then
		local numberInRaid = GetNumRaidMembers();
		if (numberInRaid > 0) then
			local myRaidGroup = nil;
			local statuses = { };
			local petStatuses = { };
			local key;
			local targetTable;

			-- initialize the arrays to empty
			for i = 1, 8, 1 do
				key = "priorityRaidGroup"..i;
				statuses[key] = { };
				petStatuses[key] = { };
			end

			-- go through the entire raid to get the appropriate healees
			for i = 1, numberInRaid do
				local raidName, _, raidGroup = GetRaidRosterInfo(i);
				local unitName;

				key = "priorityRaidGroup"..raidGroup;

				if (raidName == _SMH_myName) then
					myRaidGroup = raidGroup;
				end

				if SMH_Configuration.healingRaid or SMH_Configuration[key] then
					unitName = "raid"..i;
					if UnitIsHealable(unitName) then
						tryToInsertUnitForHealing(statuses[key], unitName, "threshold", nil);
					end
				end
				if SMH_Configuration.healingRaidPets or (SMH_Configuration[key] and SMH_Configuration.priorityRaidPets) then
					unitName = "raidpet"..i;
					if UnitIsHealable(unitName) then
						tryToInsertUnitForHealing(petStatuses[key], unitName, "petThreshold", nil);
					end
				end
			end

			-- now put the priority raid people into priority2Table and everyone else into groupStatus
			for i = 1, 8, 1 do
				local shouldProcess = true;
				key = "priorityRaidGroup"..i;

				if i == myRaidGroup then
					if not SMH_Configuration[key] then
						shouldProcess = false;
					end
				end
				if shouldProcess then
					tableSize = table.getn(statuses[key]);
					for i = 1, tableSize do
						classModifier = SMH_Configuration.classPriorities[UnitClass(statuses[key][i].id)];
						if classModifier ~= 0 then
							if SMH_Configuration[key] then tableBase = 3 else tableBase = whichPriorityLevel(classModifier, false) end
							table.insert(priority[tableBase], 1, statuses[key][i]);
						end
					end

					if SMH_Configuration[key] and SMH_Configuration.priorityRaidPets then tableBase = 3 else tableBase = 2 end
					tableSize = table.getn(petStatuses[key]);
					for i = 1, tableSize do
						table.insert(priority[tableBase], 1, petStatuses[key][i]);
					end
				end
			end
		end
	end

	-- sort each of the priority tables individually to maintain order within priority
	for i = 1, numberPriorities, 1 do
		if table.getn(priority[i]) > 0 then
			table.sort(priority[i], healthCompare);
outputCombatMessage("Priority table "..i.." has ".. table.getn(priority[i]).." healees in it.");
for j = 1, table.getn(priority[i]), i do
outputCombatMessage("Priority "..i.." order "..j.." is "..UnitName(priority[i][j].id).." "..UnitClass(priority[i][j].id).." percent "..priority[i][j].percent);
end
		end
	end

	-- and then add them to the overall table
	for i = 1, numberPriorities, 1 do
		tableSize = table.getn(priority[i]);
		if tableSize > 0 then
			for j = 1, tableSize, 1 do
				table.insert(groupStatus, j, priority[i][j]);
			end
		end
	end

	--	Put myself at front of list if my health is less than 40%
	tryToInsertUnitForHealing(groupStatus, 'player', "selfThreshold", "Player put at head of heal list");

	--	Put the target at the front of the list if it is healable and damaged
	if SMH_Configuration.targetPriority and UnitIsHealable('target') then
		local targetCurrentHealth = UnitHealth('target');
		local targetMaxHealth = UnitHealthMax('target');
		local shouldAddTarget = false;

		if (targetCurrentHealth < SMH_Configuration.threshold or UnitInParty('target') or UnitInRaid('target')) then
			-- for a target not in the party, UnitHealth returns the percentage of maximum health
			-- so we must make a guess at how much is gone based on the target's class and level
			if not UnitInParty('target') and not UnitInRaid('target') then
				local slope = (maxClassHealth[UnitClass('target')] - 30) / 59;
				targetMaxHealth = UnitLevel('target') * slope + 30;
				targetCurrentHealth = targetMaxHealth * targetCurrentHealth / 100;
				shouldAddTarget = true;
			elseif (targetCurrentHealth < targetMaxHealth) then
				shouldAddTarget = true;
			end
			if shouldAddTarget then
				local targetPercentHealth = targetCurrentHealth / targetMaxHealth;
				if (targetPercentHealth < (SMH_Configuration.threshold / 100.0)) then
					table.insert(groupStatus, 1, { id = 'target', percent = targetPercentHealth, current = targetCurrentHealth, max = targetMaxHealth });
					outputDebugMessage("The target "..UnitName('target').." put into heal list");
outputCombatMessage("The target "..UnitName('target').." put into heal list");
				end
			end
		end
	end

	--	If ALT is used, the player is put at the front of the list
	if (IsAltKeyDown() == 1) then
		tryToInsertUnitForHealing(groupStatus, 'player', "threshold", "Player put at front of heal list because of ALT");
	end

	return groupStatus;
end

--	This returns nil if there is no spell to cast.
--	The unit passed in is a table.
--	It will check all three types of spell we can cast
--	and rule out ones that cannot be cast based on what
--	the target already has on it, and what the caster
--	is doing/has on it.  It then decides on the spell
--	to cast based on the current algorithm.
local function determineSpellAndRankForUnit(groupInfo)
	local playerBuffs = unitBuffs('player');
	local casterHasNaturesSwiftness = playerBuffs[SMH_SPELL_NATURESSWIFTNESS];
	local casterHasClearcasting = playerBuffs[SMH_SPELL_CLEARCASTING];
	local targetBuffs;
	if (groupInfo.id == 'player') then
		targetBuffs = playerBuffs;
	else
		targetBuffs = unitBuffs(groupInfo.id);
	end
	local targetHasRegrowth = targetBuffs[SMH_SPELL_REGROWTH];
	local targetHasRejuvenation = targetBuffs[SMH_SPELL_REJUVENATION];
	local rankIndex = { [SMH_SPELL_REGROWTH] = 1, [SMH_SPELL_REJUVENATION] = 1, [SMH_SPELL_HEALING_TOUCH] = 1 };
	local spellsToCheck = { SMH_SPELL_REGROWTH, SMH_SPELL_REJUVENATION, SMH_SPELL_HEALING_TOUCH };
	local healAmount;
	local spellInfo;
	local trulyDesiredRank;
	local myMana = UnitMana('player');
	local spellName;
	local targetLevel = UnitLevel(groupInfo.id);
	local spellToUse = nil;
	local spellRank = nil;
	local timePenalties = { [SMH_SPELL_HEALING_TOUCH] = { 58, 43, 29, 15, 0, 0, 0, 0, 0, 0 }, [SMH_SPELL_REGROWTH] = { 21, 21, 21, 21, 21, 21, 21, 21, 21 }, [SMH_SPELL_REJUVENATION] = { 20, 20, 20, 20, 20, 20, 20, 20, 20, 20 } };

	outputDebugMessage("Determining spell to use for "..UnitName(groupInfo.id));
	if (nil == myMana) then myMana = 0; end
	for i = 1, 3 do		-- loop through each of spellsToCheck
		spellName = spellsToCheck[i];

		-- first determine the level of spell needed based on healing required
		while (true) do
			spellInfo = cachedSpellBookData[spellName][rankIndex[spellName]];
			if (spellInfo) then

				--	Determine the base heal amount garnered from the
				--	tooltip for each spell.  This includes modifications
				--	from talents, but not from equipment.  It uses the
				--	technique the user wants for healing to determine
				--	where in the range of healing the target should be.
				healAmount = spellInfo.minHeal + ((spellInfo.maxHeal - spellInfo.minHeal) * SMH_Configuration.healPercentage / 100);

				--	Now modify the base amount by the amount of extra
				--	healing equipment can do.
				if (_SMH_extraHealingAmount > 0) then
					--	Penalties are integer amounts representing percentages off of _SMH_extraHealingAmount
					local learned;
					local levelPenalty = 0;
					local timePenalty = 0;
					local totalPenalty = 0;

					--	If the spell was learned prior to level 20 we have a penalty of how much of the
					--	_SMH_extraHealingAmount that is applied.
					learned = levelLearned[spellName][rankIndex[spellName]];
					if (learned < 20) then levelPenalty = (20 - learned) * 3.75; else levelPenalty = 0; end

					--	The time penalty for a spell depends on the casting time compared to 3.5 seconds
					--	of if it is Heal Over Time, compared to 15 seconds.  These values are based on the
					--	base time for spells and ignore talent modifications.  Note that Regrowth takes
					--	2 seconds to cast for the initial spell, and the HOT is 21 seconds.  50% is given to
					--	the initial heal and 50% to the HOT.  For Rejuvenation, the HOT takes only 12 seconds
					--	so only 80% of the possible heal is permitted.
					timePenalty = timePenalties[spellName][rankIndex[spellName]];

					totalPenalty = levelPenalty + timePenalty;

--					outputDebugMessage("Initial heal amount: "..healAmount.." Raw Extra Healing: ".._SMH_extraHealingAmount);
					if (totalPenalty > 0) then
						if (totalPenalty < 100) then
							healAmount = healAmount + (_SMH_extraHealingAmount * (100 - totalPenalty) / 100);
						end
					else
						healAmount = healAmount + _SMH_extraHealingAmount;
					end
--					outputDebugMessage("Level Penalty: "..levelPenalty.."% Time Penalty: "..timePenalty.."%");
--					outputDebugMessage("Newly calculated heal amount: "..healAmount);
				end

				if (groupInfo.max - groupInfo.current < healAmount) then break end
				rankIndex[spellName] = rankIndex[spellName] + 1;
			else
				rankIndex[spellName] = rankIndex[spellName] - 1;
				break;
			end
		end

		-- now alter the rank based on overhealing
		trulyDesiredRank = rankIndex[spellName];
		local amountToAttemptToRaise = 0;
		if spellName == SMH_SPELL_HEALING_TOUCH then
			amountToAttemptToRaise = SMH_Configuration.overhealRankHealingTouch;
		elseif spellName == SMH_SPELL_REGROWTH then
			amountToAttemptToRaise = SMH_Configuration.overhealRankRegrowth;
		elseif spellName == SMH_SPELL_REJUVENATION then
			amountToAttemptToRaise = SMH_Configuration.overhealRankRejuvenation;
		end
		if amountToAttemptToRaise > 0 then
			while (cachedSpellBookData[spellName][rankIndex[spellName]] and amountToAttemptToRaise > 0) do
				rankIndex[spellName] = rankIndex[spellName] + 1;
				amountToAttemptToRaise = amountToAttemptToRaise - 1;
			end
			if not cachedSpellBookData[spellName][rankIndex[spellName]] and rankIndex[spellName] > 0 then
				rankIndex[spellName] = rankIndex[spellName] - 1;
			end
			if trulyDesiredRank ~= rankIndex[spellName] then
				outputDebugMessage("For "..spellName.." we really wanted rank "..trulyDesiredRank.." but overheal raised that to rank "..rankIndex[spellName]);
			end
		end

		-- now check to ensure we have enough mana to cast the desired spell
		trulyDesiredRank = rankIndex[spellName];
		if not casterHasClearcasting then
			while (rankIndex[spellName] > 0) do
				if (myMana >= cachedSpellBookData[spellName][rankIndex[spellName]].mana) then break end
				rankIndex[spellName] = rankIndex[spellName] - 1;
			end
			if (trulyDesiredRank ~= rankIndex[spellName]) then
				outputDebugMessage("For "..spellName.." decided we want rank "..trulyDesiredRank.." buy mana limits us to rank "..rankIndex[spellName]);
			end

		-- now we see if we have Clearcasting and if so raise the rank as high as it can go
		else
			while (cachedSpellBookData[spellName][rankIndex[spellName]]) do
				rankIndex[spellName] = rankIndex[spellName] + 1;
			end
			rankIndex[spellName] = rankIndex[spellName] - 1;	-- get back to where we have a spell
			if (trulyDesiredRank ~= rankIndex[spellName]) then
				outputDebugMessage("For "..spellName.." decided we want rank "..trulyDesiredRank.." but Clearcasting allows us to use our highest rank "..rankIndex[spellName]);
			end
		end

		-- now check to ensure spell can be cast on level of target
		trulyDesiredRank = rankIndex[spellName];
		if (rankIndex[spellName] > 0) then
			while (rankIndex[spellName] > 0) do
				if (targetLevel + 10 >= cachedSpellTargetLevels[spellName][rankIndex[spellName]]) then break end
				rankIndex[spellName] = rankIndex[spellName] - 1;
			end
			if (trulyDesiredRank ~= rankIndex[spellName]) then
				outputDebugMessage("For "..spellName.." "..UnitName(groupInfo.id).." is only level "..targetLevel.." so had to reduce rank from "..trulyDesiredRank.." to rank "..rankIndex[spellName]);
			end
		end

		outputDebugMessage("For "..spellName.." final rank is "..rankIndex[spellName]);
	end

	-- if we are moving and do not have Nature's Swiftness we are limited to Rejuvenation
	if (_SMH_isMoving and not casterHasNaturesSwiftness) then
		rankIndex[SMH_SPELL_REGROWTH] = 0;
		rankIndex[SMH_SPELL_HEALING_TOUCH] = 0;
		outputDebugMessage("We are moving without Nature's Swiftness so are limited to Rejuvenation");
	end

	-- eliminate spells the target already has
	if targetHasRegrowth then
		rankIndex[SMH_SPELL_REGROWTH] = 0;
		outputDebugMessage(UnitName(groupInfo.id).." already has Regrowth");
	end
	if targetHasRejuvenation then
		rankIndex[SMH_SPELL_REJUVENATION] = 0;
		outputDebugMessage(UnitName(groupInfo.id).." already has Rejuvenation");
	end

	-- limit ourselves to any type of spell we are forcing ourselves to use
	if healingMode == 1 then
		rankIndex[SMH_SPELL_REGROWTH] = 0;
		rankIndex[SMH_SPELL_REJUVENATION] = 0;
		outputDebugMessage("We are limiting ourselves to Healing Touch because of command.");
	elseif healingMode == 2 then
		rankIndex[SMH_SPELL_HEALING_TOUCH] = 0;
		rankIndex[SMH_SPELL_REJUVENATION] = 0;
		outputDebugMessage("We are limiting ourselves to Regrowth because of command.");
	elseif healingMode == 3 then
		rankIndex[SMH_SPELL_HEALING_TOUCH] = 0;
		rankIndex[SMH_SPELL_REGROWTH] = 0;
		outputDebugMessage("We are limiting ourselves to Rejuvenation because of command.");
	end

	-- now we use our algorithm to determine the spell to cast
	local temp = { };
	for n in pairs(SMH_Configuration.algorithm) do table.insert(temp, n) end
	table.sort(temp, greaterCompare);
	for i, n in ipairs(temp) do 
		if (groupInfo.percent * 100) > tonumber(n) then
			outputDebugMessage(string.format("the groupinfo percentage is %d and the table shows %s...a match", groupInfo.percent * 100, n));
			if rankIndex[SMH_Configuration.algorithm[n]] > 0 then
				spellToUse = SMH_Configuration.algorithm[n];
				outputDebugMessage(string.format("we have found the spell to use in table %s %s", n, spellToUse));
			end
			break;
		end
	end

	-- if no spell has been found using the algorithm do into our default spell list
	if (spellToUse == nil) then
		local tableSize = table.getn(SMH_Configuration.defaultSpells);
		for i = 1, tableSize do
			if rankIndex[SMH_Configuration.defaultSpells[i]] > 0 then
				spellToUse = SMH_Configuration.defaultSpells[i];
				break;
			end
		end
	end

	if (spellToUse) then
		spellRank = rankIndex[spellToUse];
		outputDebugMessage("Decided on "..spellToUse.." Rank("..spellRank..")");
	else
		outputDebugMessage("Found no spell to use");
	end

	return spellToUse, spellRank
end

local function outputPartyMessage(message)
	if SMH_Configuration.messagesParty then
		if UnitInParty('party1') then
			SendChatMessage(message, "PARTY")
		else
			outputChatMessage(message);
		end
	else
		outputChatMessage(message);
	end
end

--	This function walks through the spell book looking for spells in which we have interest.
--	It then caches the entries into the spell book into tables so we can easily determine
--	which spell to cast based on the name of the spell and the desired casting rank.
function smh_cacheSpellBookData()

	outputDebugMessage("Caching spell book data.");

	-- clear the cache
	cachedSpellBookData = { [SMH_SPELL_REGROWTH] = { }, [SMH_SPELL_HEALING_TOUCH] = { }, [SMH_SPELL_REJUVENATION] = { }};

	local index = 1;
	local spellsOfInterest = { [SMH_SPELL_REGROWTH] = true, [SMH_SPELL_HEALING_TOUCH] = true, [SMH_SPELL_REJUVENATION] = true };

	while (true) do
		local spellName, spellRankName = GetSpellName(index, SpellBookFrame.bookType);

		-- If there is no spellName then we have run out of spells in the spell book
		if (not spellName) then
			do break end
		end

		-- If the spellName is in our table we have interest in processing it
		if (spellsOfInterest[spellName]) then
			local _, _, spellRank = string.find(spellRankName, SMH_SPELLBOOK_RANK_FORMAT);
			local rankIndex = tonumber(spellRank);
			--	Now we look at the tooltips for each of the spells to pull out the mana cost
			--	and the amount of healing that can be done.  These values have been adjusted
			--	for the talents already.
			smhTooltip:SetSpell(index, SpellBookFrame.bookType);
			local manaLine = smhTooltipTextLeft2:GetText();
			local _,_,manaCost = string.find(smhTooltipTextLeft2:GetText(), SMH_SPELLBOOK_MANA_FORMAT);
			local maxHealAmount = spellParsersMaxHeal[_SMH_class][spellName](smhTooltipTextLeft4:GetText());
			local minHealAmount = spellParsersMinHeal[_SMH_class][spellName](smhTooltipTextLeft4:GetText());
			local entry = { id = index, mana = tonumber(manaCost), maxHeal = maxHealAmount, minHeal = minHealAmount };
			cachedSpellBookData[spellName][rankIndex] = entry;
		end

		-- Next spell please
		index = index + 1
	end
end

--	The purpose of this function is to look through all the equipped items
--	and determine how much extra healing is provided by them.
function smh_cacheEquipmentData()

	_SMH_extraHealingAmount = 0;

	for i, slot in equipmentSlots do
		id, _ = GetInventorySlotInfo(slot.."Slot");
		item, _, _ = smhTooltip:SetInventoryItem('player', id);
		if item then
			local lineCount = smhTooltip:NumLines();
			for j=1, lineCount, 1 do
				local text = getglobal("smhTooltipTextLeft"..j);
				if text then
					local line = text:GetText();

					for k, pattern in _smh_healingPatterns do
						local match, _, amount = string.find(line, pattern);
						if (match ~= nil) then
							outputDebugMessage("On "..slot.." found "..tonumber(amount).." healing.");
							_SMH_extraHealingAmount = _SMH_extraHealingAmount + tonumber(amount);
							break;
						end
					end
				end
			end
		end
	end

	outputDebugMessage("Total extra healing found on equipment: ".._SMH_extraHealingAmount);
	com_mithrandir_smhEquipmentHealing:SetText(string.format(SMH_TOTAL_EXTRA_HEALING_FORMAT, _SMH_extraHealingAmount));

end

local function canCastNaturesSwiftness()
	local retval = false;
	local start, duration, enable = GetActionCooldown(_SMH_swiftNatureSlot);
	local bogusAddition = start + duration;
	if (bogusAddition == 0) then
		retval = true;
	else
		retval = false;
	end
	return retval;
end

local function smh_healCommand(healingMode)
	if _SMH_swiftNatureSlot ~= nil and SMH_Configuration["checkingNaturesSwiftness"] and not hasBuff('player', SMH_SPELL_NATURESSWIFTNESS) and canCastNaturesSwiftness() then
		outputChatMessage(SMH_INFORMATIVE_MESSAGE_CAST_NS);
		UseAction(_SMH_swiftNatureSlot);
		return;
	end
	local groupStatus = orderedHealableUnits();
	if (table.getn(groupStatus) == 0) then
		outputChatMessage(SMH_INFORMATIVE_MESSAGE_NO_ONE);
	else
		if SpellIsTargeting() then SpellStopTargeting() end
		for key, value in groupStatus do
			local spellName, spellRank = determineSpellAndRankForUnit(value);
			local shouldBreak = false;
			if (spellName) then
				local spellId = cachedSpellBookData[spellName][spellRank].id;
				if (spellId ~= nil) then
					if (UnitIsVisible(value.id)) then
						local targetBuffs = unitBuffs(value.id);
						if not targetBuffs[SMH_SPELL_PROWL] and not targetBuffs[SMH_SPELL_STEALTH] and not targetBuffs[SMH_SPELL_SHADOWMELD] then

							-- Since we want to find the range we will set the target to the
							-- one we want to heal.  Therefore, we need to know whether we
							-- already had a target.
							local hadTarget = false;
							local targetName = nil;
							if (UnitExists('target')) then
								hadTarget = true;
								if (UnitIsFriend('target', 'player')) then targetName = UnitName('target'); end
							end
							TargetUnit(value.id);

							if (UnitIsUnit('target', value.id)) then
								if (_SMH_healingSlot == nil or IsActionInRange(_SMH_healingSlot) == 1) then
									outputPartyMessage(string.format(SMH_INFORMATIVE_MESSAGE_HEALINGFORMAT, spellName, spellRank, UnitName(value.id)));
									CastSpell(spellId, SpellBookFrame.bookType);
									shouldBreak = true;
								else
									outputChatMessage(string.format(SMH_INFORMATIVE_MESSAGE_OORFORMAT, UnitName(value.id)));
								end
							else
								outputChatMessage(string.format(SMH_INFORMATIVE_MESSAGE_CANNOT_TARGETFORMAT, UnitName(value.id)));
							end

							-- Since we actually set the target to the one we wanted to heal
							-- to be able to determine whether they are in range of our spell
							-- we have to reset the target to what it was before.  Note that
							-- the following line does not work.
							-- if (hadTarget) then TargetLastTarget(); else ClearTarget(); end
							if (hadTarget) then
								if (targetName ~= nil) then TargetByName(targetName); else TargetLastEnemy(); end
							else
								ClearTarget();
							end
							if shouldBreak then break end
						else
							outputChatMessage(string.format(SMH_INFORMATIVE_MESSAGE_STEALTHFORMAT, UnitName(value.id)));
						end
					else
						outputChatMessage(string.format(SMH_INFORMATIVE_MESSAGE_INVISIBLEFORMAT, UnitName(value.id)));
					end
				else
					outputChatMessage(string.format(SMH_INFORMATIVE_MESSAGE_SPELL_LOSTFORMAT, spellName, spellRank, UnitName(value.id)));
				end
				if SpellIsTargeting() then SpellStopTargeting() end
			else
				outputChatMessage(string.format(SMH_INFORMATIVE_MESSAGE_NO_SPELLFORMAT, UnitName(value.id)));
			end
		end
	end
end

local function smh_healHelp()
	local helpCount = table.getn(SMH_HELP_LINES);
	for i = 1, helpCount, 1 do
		outputChatMessage(string.format(SMH_HELP_LINES[i], SMH_MACRO_COMMAND));
	end
end

local outputSettingMessages = {
	["mesagesParty"] = SMH_OUTPUT_MESSAGE_MESSAGE,
	["debug"] = SMH_OUTPUT_MESSAGE_DEBUG,
	["checkingNaturesSwiftness"] = SMH_OUTPUT_MESSAGE_NATURESSWIFTNESS,
	};

local function smh_healBoolean(parameterName, yn)
	SMH_Configuration[parameterName] = yn;
	outputChatMessage(string.format("<%s>: %s %s", SMH_VERSION_STRING, outputSettingMessages[parameterName], outputBooleanMappings[yn]));
end

local function smhHealHealingTouch()
	smh_healCommand(1);
end

local function smhHealRegrowth()
	smh_healCommand(2);
end

local function smhHealRejuvenation()
	smh_healCommand(3);
end

local function smhHealNormal()
	smh_healCommand(0);
end

--	This routine is only used for testing different things
--	out and has no real purpose other than allowing me to
--	have a test harness available relatively easily.
local function smhHealTest()
end

function SMH_ShowHideOptionsUI()
	if (com_mithrandir_smhOptions:IsVisible()) then
		com_mithrandir_smhOptions:Hide();
	else
		com_mithrandir_smhOptions:Show();
	end
end

local commandFunctions = {
	[SMH_COMMAND_HELP] = smh_healHelp,
	[SMH_COMMAND_UI] = SMH_ShowHideOptionsUI,
	[SMH_COMMAND_HEALINGTOUCH] = smhHealHealingTouch,
	[SMH_COMMAND_REGROWTH] = smhHealRegrowth,
	[SMH_COMMAND_REJUVENATION] = smhHealRejuvenation,
	[SMH_COMMAND_TEST] = smhHealTest,
	};

local commandParameters = {
	[SMH_COMMAND_MSG] = "messagesParty",
	[SMH_COMMAND_DEBUG] = "debug",
	[SMH_COMMAND_NATURE] = "checkingNaturesSwiftness",
	};

function smhOnSlashCommand(msg)		-- all the parameters will be passed in the msg...therefore, we will need to split some out
	if (msg == nil or msg == "") then
		smhHealNormal();
	else
		msg = string.lower(msg);
		-- hopefully this means we will have command and options filled in such as:
		-- /smh msg => command="msg" options=nil
		-- /smh msg on => command="msg" options="on"
		local _, _, command, options = string.find(msg, "^%s*(%a+)%s*(%a*)");
		if (command ~= nil) then
			if commandFunctions[command] then
				commandFunctions[command]()
			else
				if commandParameters[command] then
					local booleanValue;
					if (options == SMH_COMMAND_OPTION_ON) then booleanValue = true
					elseif (options == SMH_COMMAND_OPTION_OFF) then booleanValue = false
					else booleanValue = not SMH_Configuration[commandParameters[command]];
					end
					smh_healBoolean(commandParameters[command], booleanValue);
				end
			end
		end
	end
end

local function SMH_Slider_SetTitle(this)
	local id = this:GetID();
	local format = _smh_sliderTitleFormats[id][2];
	getglobal(this:GetName().."Text"):SetText(string.format(format, SMH_Configuration[_smh_sliderTitleFormats[id][1]]));
end

function SMH_Slider_OnShow()
	local id = this:GetID();
	getglobal(this:GetName().."High"):SetText();
	getglobal(this:GetName().."Low"):SetText();
	SMH_Slider_SetTitle(this);
	if (id == 5) then
		this:SetMinMaxValues(0, 359);
	elseif (id == 6 or id == 8) then
		this:SetMinMaxValues(0, 9);
	elseif (id == 7) then
		this:SetMinMaxValues(0, 8);
	else
		this:SetMinMaxValues(1, 100);
	end
	this:SetValueStep(1);
	this:SetValue(SMH_Configuration[_smh_sliderTitleFormats[id][1]]);
end

function SMH_Slider_OnValueChanged()
	local id = this:GetID();
	SMH_Configuration[_smh_sliderTitleFormats[id][1]] = this:GetValue();
	SMH_Slider_SetTitle(this);
	if (id == 5) then
		com_mithrandir_smhMiniButton:SetPoint("TOPLEFT", "Minimap", "TOPLEFT", 52 - (80 * cos(this:GetValue())), (80 * sin(this:GetValue())) - 52);
	end
end

function SMH_CheckButton_OnClick()
	local key = _smh_checkButtonTitles[this:GetID()][1];

	SMH_Configuration[key] = (this:GetChecked() == 1);
end

function SMH_CheckButton_OnLoad()
	getglobal(this:GetName().."Text"):SetText(_smh_checkButtonTitles[this:GetID()][2]);
end

function SMH_CheckButton_OnShow()
	if SMH_Configuration[_smh_checkButtonTitles[this:GetID()][1]] then
		this:SetChecked(1);
	else
		this:SetChecked(0);
	end
end

function SMH_UpdateClassFrame(frame, number)
	local key = getglobal(frame:GetName().."Title"):GetText();
	getglobal(frame:GetName().."Button0"):SetChecked(0);
	getglobal(frame:GetName().."Button1"):SetChecked(0);
	getglobal(frame:GetName().."Button2"):SetChecked(0);
	getglobal(frame:GetName().."Button3"):SetChecked(0);
	getglobal(frame:GetName().."Button"..number):SetChecked(1);
	SMH_Configuration.classPriorities[key] = number;
end

function SMH_ClassFrame_OnShow()
	SMH_UpdateClassFrame(this, SMH_Configuration.classPriorities[getglobal(this:GetName().."Title"):GetText()])
end

--	This routine makes sure only the one clicked is considered checked
--	and makes sure the configuration information is updated with the
--	proper numeric value.
function SMH_ClassRadio_OnClick()
	SMH_UpdateClassFrame(this:GetParent(), this:GetID());
end

--	If forcing it means we overwrite whatever was stored
function SMH_UpdateFromDefaults(forcing)
	--	Ensure we have the basic set of default options in case the user edited the saved file
	--	or we have changed what the set is.
	for key in pairs(SMH_Configuration_Defaults) do
		if (nil == SMH_Configuration[key] or forcing) then
			SMH_Configuration[key] = SMH_Configuration_Defaults[key];
		end
	end
	--	Remove values from the configuration that do not have keys in the defaults since this
	--	means the key is no longer valid or the user has added something extra.
	for key in pairs(SMH_Configuration) do
		if (nil == SMH_Configuration_Defaults[key])  then
			SMH_Configuration[key] = nil;
		end
	end
end

local function SMH_UpdateDefaultSpellButtons()
	for i = 1, 3, 1 do
		getglobal("com_mithrandir_smhDefault"..i).tooltiptext = SMH_Configuration.defaultSpells[i];
		getglobal("com_mithrandir_smhDefault"..i):SetNormalTexture(_smh_defaultSpellIcons[SMH_Configuration.defaultSpells[i]]);
	end
end

function SMH_SwapDefaults(left, right)
	local temp = SMH_Configuration.defaultSpells[left];
	SMH_Configuration.defaultSpells[left] = SMH_Configuration.defaultSpells[right];
	SMH_Configuration.defaultSpells[right] = temp;
	SMH_UpdateDefaultSpellButtons();
end

local function smh_init()
	outputChatMessage(string.format(SMH_IntroductionTextFormat, SMH_VERSION_STRING, SMH_MACRO_COMMAND));

	SlashCmdList["SMARTHEAL"] = function(msg)
		smhOnSlashCommand(msg)
	end
	SLASH_SMARTHEAL1 = SMH_MACRO_COMMAND;

	--	Ensure our tooltip is not messed up.
	if not smhTooltip:IsOwned(UIParent) then
		smhTooltip:SetOwner(UIParent, "ANCHOR_NONE");
	end

	SMH_UpdateFromDefaults(false);

	com_mithrandir_smhMiniButton:SetPoint("TOPLEFT", "Minimap", "TOPLEFT", 52 - (80 * cos(SMH_Configuration.minimapButtonAngle)), (80 * sin(SMH_Configuration.minimapButtonAngle)) - 52);

	com_mithrandir_smhOptionsTitle:SetText("SMart Heal "..SMH_VERSION.." Options");

	SMH_UpdateDefaultSpellButtons();

	this:RegisterEvent("PLAYER_ENTERING_WORLD");
	this:RegisterEvent("PLAYER_LEAVING_WORLD");
end

function smh_onLoad()
	x, _SMH_class = UnitClass('player');
	_SMH_myName = UnitName('player');
	if (_SMH_class == SMH_CLASS_DRUID_ENGLISH) then
		this:RegisterEvent("VARIABLES_LOADED");
	else
		com_mithrandir_smhMiniButton:Hide();
	end
end

local function SMH_FrameUpdate()
	local numberTabs = getn(_smh_optionTabs);
	for i = 1, numberTabs, 1 do
		getglobal("com_mithrandir_smhTab" .. i):Show();
	end
end

local uiElements = { };

function SMH_SetTab(tabId)
	if (tabId == 4) then
		SMH_priorityString();
		SMH_algorithmString();
	end
	local numberUIElements = getn(uiElements);
	for i = 1, numberUIElements, 1 do
		if tabId == uiElements[i][2] then
			uiElements[i][1]:Show();
		else
			uiElements[i][1]:Hide();
		end
	end
end

function SMH_Tab_OnClick(tabId)
	if not tabId then
		tabId = this:GetID();
	end
	SMH_SetTab(tabId);
	local numberTabs = getn(_smh_optionTabs);
	for i = 1, numberTabs, 1 do
		if (i ~= tabId) then
			getglobal("com_mithrandir_smhTab" .. i):SetChecked(nil);
		else
			getglobal("com_mithrandir_smhTab" .. i):SetChecked(1);
		end
	end
	SMH_FrameUpdate();
end

function com_mithrandir_smhOptionsOnLoad()
	if (_SMH_class == SMH_CLASS_DRUID_ENGLISH) then
		-- Currently we set up the uiElements by hand because they are not
		-- loaded until just before we get this call.
		uiElements = {
			{ com_mithrandir_smhMessagesParty, 1 },
			{ com_mithrandir_smhUsesTarget, 1 },
			{ com_mithrandir_smhHealsParty, 1 },
			{ com_mithrandir_smhHealsPets, 1 },
			{ com_mithrandir_smhHealingRaid, 3 },
			{ com_mithrandir_smhHealingRaidPets, 3 },
			{ com_mithrandir_smhPartyPriority, 2 },
			{ com_mithrandir_smhNaturesSwiftness, 1 },
			{ com_mithrandir_smhDamageThreshold, 2 },
			{ com_mithrandir_smhPetDamageThreshold, 2 },
			{ com_mithrandir_smhSelfDamageThreshold, 2 },
			{ com_mithrandir_smhHealPercentage, 2 },
			{ com_mithrandir_smhDebug, 1 },
			{ com_mithrandir_smhMiniButtonAngle, 1 },
			{ com_mithrandir_smhReset, 1 },
			{ com_mithrandir_smhRaidInstruct, 3 },
			{ com_mithrandir_smhPriRaid1, 3 },
			{ com_mithrandir_smhPriRaid2, 3 },
			{ com_mithrandir_smhPriRaid3, 3 },
			{ com_mithrandir_smhPriRaid4, 3 },
			{ com_mithrandir_smhPriRaid5, 3 },
			{ com_mithrandir_smhPriRaid6, 3 },
			{ com_mithrandir_smhPriRaid7, 3 },
			{ com_mithrandir_smhPriRaid8, 3 },
			{ com_mithrandir_smhPriRaidPets, 3 },
			{ com_mithrandir_smhDefaultInstruct, 2 },
			{ com_mithrandir_smhDefault1, 2 },
			{ com_mithrandir_smhDefault2, 2 },
			{ com_mithrandir_smhDefault3, 2 },
			{ com_mithrandir_smhDefaultLeft, 2 },
			{ com_mithrandir_smhDefaultRight, 2 },
			{ com_mithrandir_smhPriorityDesc, 4 },
			{ com_mithrandir_smhAlgorithmDesc, 4 },
			{ com_mithrandir_smhClassPriority, 5 },
			{ com_mithrandir_smhClassHighPriority, 5 },
			{ com_mithrandir_smhOverhealing, 6 },
			{ com_mithrandir_smhOverhealHealingTouch, 6 },
			{ com_mithrandir_smhOverhealRegrowth, 6 },
			{ com_mithrandir_smhOverhealRejuvenation, 6 },
			{ com_mithrandir_smhPriorityDruid, 5 },
			{ com_mithrandir_smhPriorityHunter, 5 },
			{ com_mithrandir_smhPriorityMage, 5 },
			{ com_mithrandir_smhPriorityPaladin, 5 },
			{ com_mithrandir_smhPriorityPriest, 5 },
			{ com_mithrandir_smhPriorityRogue, 5 },
			{ com_mithrandir_smhPriorityShaman, 5 },
			{ com_mithrandir_smhPriorityWarlock, 5 },
			{ com_mithrandir_smhPriorityWarrior, 5 },
			};

		SMH_SetTab(1);
		com_mithrandir_smhTab1:SetChecked(1);
		local numberTabs = getn(_smh_optionTabs);
		for i = 1, numberTabs, 1 do
			getglobal("com_mithrandir_smhTab" .. i).tooltiptext = _smh_optionTabs[i][1];
			getglobal("com_mithrandir_smhTab" .. i):SetNormalTexture(_smh_optionTabs[i][2]);
			getglobal("com_mithrandir_smhTabButton" .. i).tooltiptext = _smh_optionTabs[i][1];
		end

		SMH_FrameUpdate();
	end
end

_SMH_event_dispatch['VARIABLES_LOADED'] = function()
	smh_init();
end

_SMH_event_dispatch['SPELLS_CHANGED'] = function(arg)
	if (not arg) then	-- arg is nil when new spell learned
		smh_cacheSpellBookData();
	end
end

_SMH_event_dispatch['PLAYER_ENTERING_WORLD'] = function()
	smh_cacheSpellBookData();
	findHealingSlot();
	smh_cacheEquipmentData();
	this:RegisterEvent("ACTIONBAR_SLOT_CHANGED");
	this:RegisterEvent("UNIT_INVENTORY_CHANGED");
	this:RegisterEvent("CHARACTER_POINTS_CHANGED");
	this:RegisterEvent("LEARNED_SPELL_IN_TAB");
end

_SMH_event_dispatch['PLAYER_LEAVING_WORLD'] = function()
	this:UnregisterEvent("ACTIONBAR_SLOT_CHANGED");
	this:UnregisterEvent("UNIT_INVENTORY_CHANGED");
	this:UnregisterEvent("CHARACTER_POINTS_CHANGED");
	this:UnregisterEvent("LEARNED_SPELL_IN_TAB");
end

_SMH_event_dispatch['LEARNED_SPELL_IN_TAB'] = function()
	smh_cacheSpellBookData();
end

_SMH_event_dispatch['ACTIONBAR_SLOT_CHANGED'] = function(arg1)
	--	arg1 tells us the slot that has changed
	if (_SMH_healingSlot == nil or _SMH_healingSlot == arg1 or _SMH_swiftNatureSlot == nil or _SMH_swiftNatureSlot == arg1) then
		findHealingSlot();
	end
end

_SMH_event_dispatch['UNIT_INVENTORY_CHANGED'] = function(arg1)
	--	arg1 tells us what unit had the inventory change
	if (arg1 == 'player') then
		smh_cacheEquipmentData();
	end
end

_SMH_event_dispatch['CHARACTER_POINTS_CHANGED'] = function(arg1)
	--	arg1 tells us the delta in talent points
	if (arg1 == -1) then
		smh_cacheSpellBookData();	-- because spell tooltips are modified based on talents
	end
end

function smh_onEvent()
	if _SMH_event_dispatch[event] then
		_SMH_event_dispatch[event](arg1, arg2)
	end
end

function smh_onUpdate(delay)
	if (_SMH_class == SMH_CLASS_DRUID_ENGLISH) then
		local positionX, positionY = GetPlayerMapPosition('player');
		if positionX ~= _SMH_lastX or positionY ~= _SMH_lastY then
			_SMH_isMoving = true;
		else
			_SMH_isMoving = false;
		end
		_SMH_lastX = positionX;
		_SMH_lastY = positionY;
	end
end
